
package mobicomp.net;

import java.io.IOException;
import java.util.HashMap;
import java.util.HashSet;

import mobicomp.interfaces.MHMessage;
import mobicomp.interfaces.MHSocket;
import mobicomp.interfaces.Message;
import mobicomp.interfaces.SHSocket;
import mobicomp.util.Dispatcher;

/**
 * TODO
 * 
 * @author Andreas Wetzel <andi@spamato.net>
 */
public class MHSocketImpl implements MHSocket
{
    public static final int DEFAULT_TIMEOUT = 1000;
    public static final int MAX_TTL = 10;
    
    // our lower layered singlehop socket
    private SHSocket lowerLayer;
    
    private HashMap routingTable = new HashMap();
    
    // id of the last flood packet
    private int lastFloodId = 0;
    // storage for all flood messages we have seen before
    private HashSet seenFloodMessages = new HashSet();
    
    private Thread messageReceiverThread;
    
    private Dispatcher dispatcher = new Dispatcher(); 


    public MHSocketImpl( int address) throws IOException {
        lowerLayer = new SHSocketImpl( address);
        
        messageReceiverThread = new MHMessageReceiver();
        messageReceiverThread.start();
        
        dispatcher.addObserver( new RouteRequestHandler( this));
    }
    
    /**
     * @see mobicomp.interfaces.MHSocket#setOwnAddress(int)
     */
    public void setOwnAddress( int addr) {
        lowerLayer.setOwnAddress( addr);
    }

    /**
     * @see mobicomp.interfaces.MHSocket#getOwnAddress()
     */
    public int getOwnAddress() {
        return lowerLayer.getOwnAddress();
    }

    /**
     * @see mobicomp.interfaces.MHSocket#send(mobicomp.interfaces.MHMessage)
     */
    public void send( MHMessage message) throws IOException {
        if (message.getReceiver() == BROADCAST_ADDRESS) {
            // generate a floodid and send message
            message.setFloodId( nextFloodId());
            broadcastMHMessage( message);
        }
        else {
            // handle addressed messages
            
            // do we know a route to the receiver?
            // if not, try to find out...
            Integer receiverHash = new Integer( message.getReceiver());
            for ( int ttl=1; (!routingTable.containsKey( receiverHash) && (ttl <= MAX_TTL)); ttl++) {

                // broadcast for new route - assemble route request message
                MHMessage rreq = new MHMessageImpl();
                rreq.setSender( getOwnAddress());
                // we use the receiver field for the address we're looking for
                rreq.setReceiver( message.getReceiver());
                rreq.setType( MHMessage.TYPE_RREQ);
                rreq.getRoute().add( new Integer( getOwnAddress()));

                // TODO replace 10 with constant
                rreq.setTTL( ttl);
                rreq.setFloodId( nextFloodId());
                System.out.println( "flooding with ttl=" + ttl);
                broadcastMHMessage( rreq);

                synchronized (this) {
                    try {
                        this.wait( DEFAULT_TIMEOUT);
                    }
                    catch (InterruptedException e) {
                        // TODO Auto-generated catch block
                        e.printStackTrace();
                    }
                }
            }
        }
    }
    
    private void sendMHMessage( MHMessage message, int to) throws IOException {
        Message shm = new SHMessageImpl();
        shm.setSender( getOwnAddress());
        shm.setReceiver( to);
        // TODO: this cast is not nice. 
        shm.setData( ((MHMessageImpl)message).toByteArray());
        lowerLayer.send( shm);
    }

    protected void broadcastMHMessage( MHMessage message) throws IOException {
        seenFloodMessages.add( genFloodHash( message));
        sendMHMessage( message, SHSocket.BROADCAST_ADDRESS);
    }

    private int nextFloodId() {
        lastFloodId++;
        if ( lastFloodId > 0xFFFF ) {
            lastFloodId = 1;
        }
        return lastFloodId;
    }
    
    // generate a unique key object for the given MHMessage
    private Object genFloodHash( MHMessage message) {
        return new Integer( ( message.getSender() << 16) + message.getFloodId());
    }
    
    private class MHMessageReceiver extends Thread {
        
        public MHMessageReceiver() {
            
        }
        
        public void run() {
            while (true) {
                try {
                    Message received = lowerLayer.receive();
                    MHMessage mhm = new MHMessageImpl( received.getData());

                    // if floodId is not zero, it is a broadcast message
                    if (mhm.getFloodId() != 0) {
                        // check if message already seen
                        Object hash = genFloodHash( mhm);
                        if (seenFloodMessages.contains( hash)) {
                            // if yes, do nothing
                            continue;
                        }
                        else {
                            // otherwise add to seen flood messages and process message
                            seenFloodMessages.add( hash);
                        }
                    }
                    dispatcher.dispatch( mhm);
                }
                catch (IOException e) {
                    // TODO Auto-generated catch block
                    e.printStackTrace();
                }                
            }
        }
    }
}
